import asyncio
from typing import Dict
from uuid import uuid4
from concurrent.futures import ProcessPoolExecutor

from fastapi import FastAPI, Request, Response
from fastapi.responses import PlainTextResponse
from playwright.async_api import async_playwright
from playwright_recaptcha import (
    RecaptchaNotFoundError,
    RecaptchaVersionError,
    RecaptchaSolveError,
    RecaptchaError,
    recaptchav3,
    recaptchav2,
)
import os
from datetime import datetime, timedelta

app = FastAPI()
loop = asyncio.get_event_loop()
tokens: Dict[str, str] = {}

executor = ProcessPoolExecutor()

# To trigger the solving


async def solve_captcha2(URL: str, SAVEFILE: str, randID: str) -> str:
    async with async_playwright() as playwright:
        # using headless mode below
        browser = await playwright.firefox.launch(headless=True, proxy={
            #        browser = await playwright.firefox.launch(proxy={
            "server": "http://p.webshare.io:80",
            "username": "PROXY-USERNAME",
            "password": "PROXY-PASSWORD"
        })
        page = await browser.new_page()
        await page.goto(URL, wait_until="networkidle")
        print(f'v2 URL|{URL}')
        try:
            async with recaptchav2.AsyncSolver(page) as solver:
                token = await solver.solve_recaptcha()
        except RecaptchaNotFoundError:
            with open(SAVEFILE, 'w') as file:
                file.write('ERROR_CAPTCHA_UNSOLVABLE')
                file.close()
            return PlainTextResponse('ERROR_CAPTCHA_UNSOLVABLE')
        except RecaptchaVersionError:
            with open(SAVEFILE, 'w') as file:
                file.write('ERROR_CAPTCHA_UNSOLVABLE')
                file.close()
            return PlainTextResponse('ERROR_CAPTCHA_UNSOLVABLE')
        except RecaptchaSolveError:
            with open(SAVEFILE, 'w') as file:
                file.write('ERROR_CAPTCHA_UNSOLVABLE')
                file.close()
            return PlainTextResponse('ERROR_CAPTCHA_UNSOLVABLE')
        except RecaptchaError:
            with open(SAVEFILE, 'w') as file:
                file.write('ERROR_CAPTCHA_UNSOLVABLE')
                file.close()
            return PlainTextResponse('ERROR_CAPTCHA_UNSOLVABLE')

        with open(SAVEFILE, 'w') as file:
            file.write(f'OK|{token}')
            file.close()

        tokens[randID] = token

    return f'OK|{token}'


async def solve_captcha3(URL: str, SAVEFILE: str, randID: str) -> str:
    async with async_playwright() as playwright:
        # using headless mode below
        browser = await playwright.firefox.launch(headless=True)
#        browser = await playwright.firefox.launch()
        page = await browser.new_page()
        await page.goto(URL)
        print(f'v3 URL|{URL}')
        try:
            async with recaptchav3.AsyncSolver(page) as solver:
                token = await solver.solve_recaptcha()
        except RecaptchaNotFoundError:
            with open(SAVEFILE, 'w') as file:
                file.write('ERROR_CAPTCHA_UNSOLVABLE')
                file.close()
            return PlainTextResponse('ERROR_CAPTCHA_UNSOLVABLE')
        except RecaptchaVersionError:
            with open(SAVEFILE, 'w') as file:
                file.write('ERROR_CAPTCHA_UNSOLVABLE')
                file.close()
            return PlainTextResponse('ERROR_CAPTCHA_UNSOLVABLE')
        except RecaptchaSolveError:
            with open(SAVEFILE, 'w') as file:
                file.write('ERROR_CAPTCHA_UNSOLVABLE')
                file.close()
            return PlainTextResponse('ERROR_CAPTCHA_UNSOLVABLE')
        except RecaptchaError:
            with open(SAVEFILE, 'w') as file:
                file.write('ERROR_CAPTCHA_UNSOLVABLE')
                file.close()
            return PlainTextResponse('ERROR_CAPTCHA_UNSOLVABLE')

        with open(SAVEFILE, 'w') as file:
            file.write(f'OK|{token}')
            file.close()

        tokens[randID] = token

    return f'OK|{token}'

# Background Async loop


def background_captcha3_solve(URL: str, SAVEFILE: str, randID: str) -> None:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        result = loop.run_until_complete(solve_captcha3(URL, SAVEFILE, randID))
        if result.startswith('ERROR_'):
            tokens[randID] = result
    except Exception as e:
        with open(SAVEFILE, 'w') as file:
            file.write('ERROR_CAPTCHA_UNSOLVABLE')
            file.close()
        tokens[randID] = 'ERROR_CAPTCHA_UNSOLVABLE'
        print(f'Error solving v3 captcha: {e}')


def background_captcha2_solve(URL: str, SAVEFILE: str, randID: str) -> None:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        result = loop.run_until_complete(solve_captcha2(URL, SAVEFILE, randID))
        if result.startswith('ERROR_'):
            tokens[randID] = result
    except Exception as e:
        with open(SAVEFILE, 'w') as file:
            file.write('ERROR_CAPTCHA_UNSOLVABLE')
            file.close()
        tokens[randID] = 'ERROR_CAPTCHA_UNSOLVABLE'
        print(f'Error solving v2 captcha: {e}')


# in.app endpoint to receive captcha parameters


@app.post("/in.php")
async def captcha2(request: Request):
    delete_old_files()
    form = await request.form()
    pageurl = form.get("pageurl")
    version = form.get("version")

    randID = str(uuid4())
    SAVEFILE = randID + '.txt'
    with open(SAVEFILE, 'w') as file:
        file.write('CAPCHA_NOT_READY')
        file.close()

    tokens[randID] = 'CAPCHA_NOT_READY'

    if version == 'v3':
        loop.run_in_executor(executor, background_captcha3_solve, pageurl, SAVEFILE, randID)

    elif version != 'v3':
        loop.run_in_executor(executor, background_captcha2_solve, pageurl, SAVEFILE, randID)

    if tokens[randID] == 'ERROR_CAPTCHA_UNSOLVABLE':
        strID = 'ERROR_CAPTCHA_UNSOLVABLE'
    else:
        strID = f'OK|{randID}'

    return PlainTextResponse(content=strID)


# res.app endpoint to respond with captcha status and solving


@app.get("/res.php")
async def getrespones(request: Request):
    action = request.query_params.get("action")
    capID = request.query_params.get("id")

    if action == 'getbalance':
        return PlainTextResponse('100')
    elif action == 'get':
        while True:
            file = open((capID)+'.txt', 'r')
            line = file.readline()
            if line != '':
                return PlainTextResponse(line)
            file.close()
    elif action == 'reportgood':
        #        print('\x1b[32m' + 'Another ReCaptchav3 Success' + '\x1b[0m')
        return PlainTextResponse('OK_REPORT_RECORDED')

# Delete txt files to clean up status and solving files older then 30 mins


def delete_old_files():
    # Set the threshold time to 30 minutes ago
    threshold_time = datetime.now() - timedelta(minutes=30)

    # Loop through all the files in the directory
    for file_name in os.listdir():
        # Check if the file is a .txt file and if its creation time is older than the threshold time
        if file_name.endswith('.txt') and datetime.fromtimestamp(os.path.getctime(file_name)) < threshold_time:
            os.remove(file_name)


if __name__ == '__main__':
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8000)
